#include "BrowseWindow.h"
#include "DiagnosticsPanel.h"
#include "GetText.h"

@implementation BrowseWindow

#define MINCOL 125
#define WIDTH  (MINCOL*3)
#define HEIGHT 300

static int count = 0;
- initWithController:(Controller *)control
{
    NSNotificationCenter *ncenter = 
        [NSNotificationCenter defaultCenter];
    NSRect browseRect = {{0, 0}, {WIDTH, HEIGHT}};
    NSRect winRect = 
    {{250+(count%12)*24, 100+(count%12)*24}, {WIDTH, HEIGHT}};
    unsigned int style = NSTitledWindowMask | NSClosableWindowMask |
        NSMiniaturizableWindowMask | NSResizableWindowMask;
    NSString *title =              
        [NSString stringWithFormat:@"GSMBrowser. # %d", ++count];              

    [self initWithContentRect:winRect
          styleMask:style
          backing:NSBackingStoreRetained
          defer:NO];
    [self setMinSize:NSMakeSize(WIDTH, HEIGHT)];
    [self setReleasedWhenClosed:YES];

    [ncenter addObserver:self 
             selector:@selector(columnLoaded:) 
             name:@"ColumnLoaded"
             object:self];
   
    browser = [[NSBrowser alloc] initWithFrame:browseRect];
    [browser setAutoresizingMask: NSViewHeightSizable | NSViewWidthSizable];
    [browser setDelegate:self];
    [browser setAllowsMultipleSelection:NO];
    [browser setMinColumnWidth:MINCOL];
    
    [self setContentView:browser];

    err = nil;
    srv = nil;

    [self setDelegate:control];
    con = control;

    [self setTitle:title];
    [self display];
    [self makeKeyAndOrderFront:nil]; // loads column zero

    return self;
}

- (NSBrowser *)browser
{
    return browser;
}

- (NSString *)error
{
    return err;
}

- (NSString *)domain
{
    return
        [[browser selectedCellInColumn:0]
            stringValue];
}

- (NSString *)fullShare
{
    return
        [NSString stringWithFormat:@"//%@/%@",
                  [[browser selectedCellInColumn:1]
                      stringValue],
                  [[browser selectedCellInColumn:2]
                      stringValue]];
}



- (void)columnLoaded:(NSNotification*)notification
{
    int column = 
        [(NSNumber *)
            [[notification userInfo] objectForKey:@"Column"]
            intValue];

    if(err!=nil){
        NSRunAlertPanel(@"Error", @"%s", @"Ok", NULL, NULL,
            [err cString]);
        [err release];
        if(!column){
            [self close];
        }
        return;
    }
}

- (NSMutableSet *)getDomains:(NSEnumerator *)cmdOutput
{
    NSMutableArray *ips = 
        [NSMutableArray arrayWithCapacity:1];
    NSMutableSet *result =
        [NSMutableSet setWithCapacity:1];

    NSString *line = nil;
    const char *servstart, *servend;

    last = @"no output";
    if((line = [cmdOutput nextObject])==nil){
        return nil;
    }

    last = 
        [NSString stringWithFormat:
                      @"wrong prefix on first line: %@", line];
    if([line hasPrefix:@"querying *"]==NO){
        return nil;
    }

    while((line = [cmdOutput nextObject])!=nil){
        const char *cstr = [line cString];
        int len = [line length];
        
        if(len>=4+4+5 && !strcmp(cstr+len-5-1, " *<00>")){
            int ipb[4];
            sscanf(cstr, "%d.%d.%d.%d ", 
                   ipb, ipb+1, ipb+2, ipb+3);
            NSString *ip =
                [NSString stringWithFormat:@"%d.%d.%d.%d",
                          ipb[0], ipb[1], ipb[2], ipb[3]];

            // verify IP
            if([line 
                   hasPrefix:
                       [ip stringByAppendingString:@" "]]==YES){
                [ips addObject:ip];
            }
        }
        last = line;
    }

    if(![ips count]){
        return nil;
    }

    NSString *ip;
    NSEnumerator *ipEnum = [ips objectEnumerator];

    while((ip = [ipEnum nextObject])!=nil){
        NSArray *args =
            [NSArray arrayWithObjects:@"-A", ip, nil];
        NSEnumerator *ipOutput = 
            [con runCommand:[con nmblookup] arguments:args 
                 input:nil progress:NO];

        NSString *ipLine = [ipOutput nextObject];
        if(ipLine==nil || 
           [ipLine hasPrefix:@"Looking up status of"]==NO){
            continue;
        }

        while((ipLine = [ipOutput nextObject])!=nil){
            const char *cstr = [ipLine cString];
            if(strstr(cstr, "<00> - <GROUP> M <ACTIVE>")
               !=NULL){
                char buf[1+[ipLine length]];
                int len;
                sscanf(cstr, "%s", buf);

                if(len=strlen(buf)){
                    NSString *dom =
                        [NSString stringWithCString:buf
                                  length:len];
                    [result addObject:dom];
                }
            }
            last = ipLine;
        }
    }

    if(![result count]){
        return nil;
    }

    return result;
}

- (NSString *)getServer:(NSEnumerator *)cmdOutput
{
    NSString *line = nil;
    const char *servstart, *servend;

    last = @"no output";
    while((line = [cmdOutput nextObject])!=nil){
        if(strstr([line cString], 
                  "<00> -         M <ACTIVE>")!=NULL){
            break;
        }
        last = line;
    }
    if(line==nil){
        return nil;
    }

    servstart = [line cString];
    while(isspace(*servstart)){
        servstart++;
    }
    servend = servstart;
    while(!isspace(*servend++));

    return 
        [NSString stringWithCString:servstart
                  length:servend-servstart-1];
}

typedef enum {
    S_PREFIX = 0,
    S_DATHEADER,
    S_DATA,
    S_DONE
} S_STATE;

- (NSMutableArray *)getHosts:(NSEnumerator *)cmdOutput
{
    NSString *line = nil, *host;
    const char *hoststart, *hostend;
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:1];
    S_STATE state = S_PREFIX;

    last = @"no output";
    while((line = [cmdOutput nextObject])!=nil){
        switch(state){
            case S_PREFIX:
                if(strstr([line cString],
                          "Server               Comment")!=NULL){
                    state = S_DATHEADER;
                }
                break;
            case S_DATHEADER:
                if(strstr([line cString],
                          "---------            -------")==NULL){
                    return nil;
                }
                state = S_DATA;
                break;
            case S_DATA:
                hoststart = [line cString];
                if(!(*hoststart)){
                    state = S_DONE;
                }
                else{
                    hoststart++;
                    hostend = hoststart;
                    while(!isspace(*hostend++));

                    host = [NSString stringWithCString:hoststart
                                     length:hostend-hoststart-1];
                    [array addObject:host];
                }
        }
        if(state==S_DONE){
            break;
        }
        last = line;
    }

    if(![array count]){
        return nil;
    }

    return array;
}

typedef struct {
    NSString *name;
    BOOL isDisk;
} ShareStruct;

int cmpshares(id a, id b, void *context)
{
    ShareStruct sha, shb;
    int c;

    [a getValue:&sha]; [b getValue:&shb];

    if(sha.isDisk!=shb.isDisk){
        return (sha.isDisk==YES ?
                NSOrderedAscending : NSOrderedDescending);
    }

    if(!(c = strcasecmp([sha.name cString], [shb.name cString]))){
        return NSOrderedSame;
    }

    return (c<0 ? NSOrderedAscending : NSOrderedDescending);
}

- (NSMutableArray *)getShares:(NSEnumerator *)cmdOutput
{
    NSString *line = nil;
    ShareStruct share;
    NSValue *val;
    const char *sharestart, *shareend, *start;
    BOOL isDisk;
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:1];
    S_STATE state = S_PREFIX;

    last = @"no output";
    while((line = [cmdOutput nextObject])!=nil){
        switch(state){
            case S_PREFIX:
                if(strstr([line cString],
                          "Sharename      Type      Comment")!=NULL){
                    state = S_DATHEADER;
                }
                break;
            case S_DATHEADER:
                if(strstr([line cString],
                          "---------      ----      -------")==NULL){
                    return nil;
                }
                state = S_DATA;
                break;
            case S_DATA:
                sharestart = start = [line cString];
                if(!(*sharestart)){
                    state = S_DONE;
                }
                else{
                    sharestart++;
                    shareend = sharestart;
                    while(!isspace(*shareend++));

                    if(strchr(sharestart, '$')==NULL &&
                       (isDisk = YES, !strncmp(start+16, "Disk", 4)) ||
                       (isDisk = NO, !strncmp(start+16, "Printer", 7))){
                        share.name = [NSString stringWithCString:sharestart
                                               length:shareend-sharestart-1];
                        share.isDisk = isDisk;
                        val = [NSValue valueWithBytes:&share
                                       objCType:@encode(ShareStruct)];
                        [array addObject:val];
                    }
                }
        }
        if(state==S_DONE){
            break;
        }
        last = line;
    }

    if(![array count]){
        return nil;
    }

    [array sortUsingFunction:cmpshares context:NULL];
    return array;
}

- (NSMutableArray *)getFiles:(NSEnumerator *)cmdOutput
{
    NSString *line = nil, *name;
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:1];
    S_STATE state = S_PREFIX;
    const char *start, *fileStart, *fileEnd;
    BOOL isDisk, error = NO;
    ShareStruct file;
    NSValue *val;
    int len;

    last = @"no output";
    while((line = [cmdOutput nextObject])!=nil){
        start = [line cString];
        switch(state){
            case S_PREFIX:
                if(strstr(start, "ERRDOS - ERRbadpath")!=NULL ||
                   strstr(start, "ERRDOS - ERRbadfile")!=NULL ||
                   strstr(start, "ERRDOS - ERRnoaccess")!=NULL ||
                   strstr(start, "ERRHRD - ERRnotready")!=NULL ||
                   strstr(start, "NT_STATUS_ACCESS_DENIED")!=NULL ||
                   strstr(start, "NT_STATUS_OBJECT_PATH_NOT_FOUND")!=NULL ||
                   !strncmp(start, "session setup failed", 20) ||
                   !strncmp(start, "tree connect failed", 19)){
                    last = line;
                    return nil;
                }

                if(strlen(start)<5 ||
                   start[0]!=' ' || start[1]!=' ' || 
                   isspace(start[2])){
                    break;
                }
                state = S_DATA;
            case S_DATA:
                if(!(*start)){
                    state = S_DONE;
                }
                else{
                    int digits = 0;
                    len=strlen(start);
                    
                    fileEnd = start+len-24-2-1;
                    while(isdigit(*fileEnd)){
                        fileEnd--; digits++;
                    }
                    digits -= 8;

                    isDisk = NO; // smbclient magic number
                    if(len==74 || (digits>0 && len-digits==74)){
                        fileEnd = start+38;
                        if(!isspace(*fileEnd)){
                            while(!isspace(*fileEnd)){
                                if(*fileEnd=='D'){
                                    isDisk = YES;
                                }
                                fileEnd--;
                            }
                        }
                        while(isspace(*fileEnd)){
                            fileEnd--;
                        }
                        name = [NSString stringWithCString:start+2
                                         length:fileEnd-start-1];
                    }
                    else{
                        const char *perms;
                         // smbclient magic number
                        int baselen = 30, namelen;

                        NSAssert(len>74, @"directory listing too short");

                        if(digits>0){
                            len -= digits;
                        }

                        namelen = baselen+len-74;

                        perms = (start+2)+namelen;
                        while(isspace(*perms)){
                            perms++;
                        }
                        if(!isdigit(*perms)){
                            while(!isspace(*perms)){
                                if(*perms=='D'){
                                    isDisk = YES;
                                    break;
                                }
                                perms++;
                            }
                        }

                        name = [NSString stringWithCString:start+2
                                         length:namelen];
                    }

                    if(isDisk==NO ||
                       ([name isEqual:@"."]==NO &&
                        [name isEqual:@".."]==NO)){
                        file.name = name;
                        file.isDisk = isDisk;
                        val = [NSValue valueWithBytes:&file
                                       objCType:@encode(ShareStruct)];
                        [array addObject:val];
                    }
                }
        }
        if(state==S_DONE){
            break;
        }
        last = line;
    }

    [array sortUsingFunction:cmpshares context:NULL];
    return array;
}



- (NSString *)browser:(NSBrowser *)sender titleOfColumn:(int)column
{
    if(!column){
        return @"Domains";
    }

    return [[browser selectedCellInColumn:column-1] stringValue];
}

- (NSMutableArray *)cmdOptionsForColumn:(int)column
{
    NSMutableArray *args = 
        [NSMutableArray arrayWithObjects:[self fullShare],
                        @"-U", [con userAndPassword], nil];

    if(column>3){
        int col; NSString *dir = @"";

        for(col=3; col<column; col++){
            dir = [dir stringByAppendingFormat:@"/%@",
                       [[browser selectedCellInColumn:col]
                           stringValue]];
        }
            
        [args addObject:@"-D"];
        [args addObject:dir];
    }

    return args;
}


- (void)browser:(NSBrowser *)sender createRowsForColumn:(int)column
       inMatrix:(NSMatrix *)matrix
{
    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSMutableArray *hosts, *shares, *files;

    NSValue *val;
    ShareStruct share, file;

    int ind, count;
    NSBrowserCell *cell;

    NSNotification *colNotif;
    NSNotificationQueue *nqueue = 
        [NSNotificationQueue defaultQueue];

    err = nil;
    
    if(!column){
        args = [NSMutableArray arrayWithObject:@"*"];
        cmdOutput = 
            [con runCommand:[con nmblookup] arguments:args 
                 input:nil progress:NO];

        NSMutableSet *domains = [self getDomains:cmdOutput];
        if(domains==nil){
            err = [NSString stringWithFormat:@"no domains: %@",
                            last];
        }
        else{
            NSArray *sortedDomains =
                [[domains allObjects]
                    sortedArrayUsingSelector:
                        @selector(caseInsensitiveCompare:)];

            [matrix addColumn];
            for(ind=0, count = [sortedDomains count]; 
                ind<count; ind++){
                NSString *dom = 
                    [sortedDomains objectAtIndex:ind];

                if(ind){
                    [matrix addRow];
                }
                cell = [matrix cellAtRow:ind column:0];
                [cell setStringValue:dom];
                [cell setLeaf:NO];
            }
        }
    }
    else if(column==1){
        if(srv!=nil){
            [srv release];
            srv = nil;
        }

        args = [NSMutableArray arrayWithObjects:@"-S", 
                               @"-M", [self domain], nil];

        cmdOutput = 
            [con runCommand:[con nmblookup] arguments:args 
                 input:nil progress:NO];
        if((srv = [self getServer:cmdOutput])==nil){
            err = [NSString stringWithFormat:@"no server: %@\n%@",
                            [self domain], last];
        }
        else{
            [srv retain];

            args = [NSMutableArray arrayWithObjects:@"-U",
                                   [con userAndPassword], @"-L", srv, nil];
            cmdOutput = 
                [con runCommand:[con smbclient] arguments:args 
                     input:nil progress:NO];
            if((hosts = [self getHosts:cmdOutput])==nil){
                err = [NSString 
                          stringWithFormat:@"no hosts in %@, server %@\n%@",
                          [self domain], srv, last];
            }
            else{
                [matrix addColumn];
                for(ind=0, count = [hosts count]; ind<count; ind++){
                    if(ind){
                        [matrix addRow];
                    }
                    cell = [matrix cellAtRow:ind column:0];
                    [cell setStringValue:[hosts objectAtIndex:ind]];
                }
            }
        }
    }
    else if(column==2){
        NSString *host = 
            [[browser selectedCellInColumn:1] stringValue];

        args = [NSMutableArray arrayWithObjects:@"-U",
                               [con userAndPassword], @"-L", host, nil];
        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:NO];
        if((shares = [self getShares:cmdOutput])==nil){
            err = [NSString 
                      stringWithFormat:@"no shares on server %@ (%@)\n%@",
                      host, [self domain], last];
        }
        else{
            [matrix addColumn];
            for(ind=0, count = [shares count]; ind<count; ind++){
                [[shares objectAtIndex:ind] getValue:&share];

                if(ind){
                    [matrix addRow];
                }
                cell = [matrix cellAtRow:ind column:0];
                [cell setStringValue:share.name];
                [cell setLeaf:(share.isDisk==YES ? NO : YES)];
            }
        }
    }
    else{
        args = [self cmdOptionsForColumn:column];
        [args addObject:@"-c"];
        [args addObject:@"ls"];

        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:NO];
        if((files = [self getFiles:cmdOutput])==nil){
            err = [NSString 
                      stringWithFormat:
                          @"no files found; domain %@\n%@\n%@",
                      [self domain], 
                      [NSString stringWithFormat:@"/%@",
                                [browser path]],
                      last];
        }
        else if([files count]){
            [matrix addColumn];
            for(ind=0, count = [files count]; ind<count; ind++){
                [[files objectAtIndex:ind] getValue:&file];

                if(ind){
                    [matrix addRow];
                }
                cell = [matrix cellAtRow:ind column:0];
                [cell setStringValue:file.name];
                [cell setLeaf:(file.isDisk==YES ? NO : YES)];
            }
        }
    }

    if(err!=nil){
        [err retain];
    }

    colNotif = 
        [NSNotification 
            notificationWithName:@"ColumnLoaded"
            object:self
            userInfo:[NSDictionary 
                         dictionaryWithObject:
                             [NSNumber numberWithInt:column]
                         forKey:@"Column"]];
    [nqueue enqueueNotification:colNotif
            postingStyle:NSPostWhenIdle];
}


- umountModal:(id)sender
{
    [[NSApplication sharedApplication] stopModal];
    umount = [sender tag];
    return self;
}

- refresh:(id)sender
{
    [browser reloadColumn:[browser lastColumn]];
    return self;
}


- upload:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];

    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line;

    int column = [browser selectedColumn];
    if([[browser selectedCell] isLeaf]==NO){
        column++;
    }

    [openPanel 
        setTitle:
            [NSString stringWithFormat:@"Upload to /%@",
                      [browser path]]];
    [openPanel setAllowsMultipleSelection:NO];
    [openPanel setPrompt:@"File:"];
    [openPanel setCanChooseDirectories:NO];

    if([openPanel runModalForTypes:nil]==NSOKButton){
        NSString *cmd;

        cmd = [NSString stringWithFormat:@"lcd \"%@\"; ",
                        [openPanel directory]];
        cmd = [cmd stringByAppendingFormat:@"put \"%@\"",
                   [[openPanel filename] lastPathComponent]];

        args = [self cmdOptionsForColumn:column];
        [args addObject:@"-c"];
        [args addObject:cmd];


        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:YES];

        while(line =[cmdOutput nextObject]){
            last = line;
        }

        if([last hasPrefix:@"putting file"]==NO){
            NSRunAlertPanel(@"Upload failed", @"/%@\n%@", 
                            @"Ok", NULL, NULL,
                            [browser path], last);
        }
        else{
            [browser reloadColumn:column];
            [browser scrollColumnToVisible:column];
        }
    }

    return self;
}

- download:(id)sender
{
    NSSavePanel *savePanel = [NSSavePanel savePanel];
    NSString *path = 
        [[NSFileManager defaultManager] currentDirectoryPath];
    
    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line, *item = [[browser selectedCell] stringValue];

    int column = [browser selectedColumn];

    [savePanel 
        setTitle:
            [NSString stringWithFormat:@"Download /%@",
                      [browser path]]];
    [savePanel setPrompt:@"File:"];

    if([savePanel runModalForDirectory:path file:item]==NSOKButton){
        NSString *cmd;

        cmd = [NSString stringWithFormat:@"lcd \"%@\"; ",
                        [savePanel directory]];
        cmd = [cmd stringByAppendingFormat:@"get \"%@\" \"%@\"",
                   item, [savePanel filename]];

        args = [self cmdOptionsForColumn:column];
        [args addObject:@"-c"];
        [args addObject:cmd];


        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:YES];

        while(line =[cmdOutput nextObject]){
            last = line;
        }

        if([last hasPrefix:@"getting file"]==NO){
            NSRunAlertPanel(@"Download failed", @"/%@\n%@", 
                            @"Ok", NULL, NULL,
                            [browser path], last);
        }
    }

    return self;
}

- deleteItemIsDirectory:(BOOL)isDir
{
    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line, *item = [[browser selectedCell] stringValue];

    int result, column = [browser selectedColumn];

    result = NSRunAlertPanel(@"Confirm", @"Really delete %@?",
                             @"No", @"Yes", NULL, item);
    if(result==NSAlertAlternateReturn){
        args = [self cmdOptionsForColumn:column];
        [args addObject:@"-c"];
        [args addObject:[NSString stringWithFormat:@"%@ \"%@\"", 
                                  (isDir==YES ? @"rmdir" : @"rm"), 
                                  item]];

        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:YES];

        while(line =[cmdOutput nextObject]){
            last = line;
        }

        if([last hasPrefix:@"Domain"]==NO &&
            (strstr([last cString], "ERR")!=NULL ||
             strstr([last cString], "NT_STATUS")!=NULL)){
            NSRunAlertPanel(@"Delete failed", @"/%@\n%@", 
                            @"Ok", NULL, NULL,
                            [browser path], 
                            ([last length]>80 ? 
                             [last substringToIndex:79] : last));
        }
        else{
            if(isDir){
                column--;
            }
            [browser reloadColumn:column];
            [browser scrollColumnToVisible:column];
        }
    }

    return self;
}

- delete:(id)sender
{
    return [self deleteItemIsDirectory:NO];
}

- mkdir:(id)sender
{
    NSString *name = RunPrompt(@"Make directory", @"Directory:");

    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line;

    int column = [browser selectedColumn];
    if([[browser selectedCell] isLeaf]==NO){
        column++;
    }

    if(name==nil){
        return self;
    }

    args = [self cmdOptionsForColumn:column];
    [args addObject:@"-c"];
    [args addObject:[NSString stringWithFormat:@"mkdir \"%@\"", name]];

    cmdOutput = 
        [con runCommand:[con smbclient] arguments:args 
             input:nil progress:YES];

    while(line =[cmdOutput nextObject]){
        last = line;
    }

    if([last hasPrefix:@"Domain"]==NO &&
       (strstr([last cString], "ERR")!=NULL ||
        strstr([last cString], "NT_STATUS")!=NULL)){
        NSRunAlertPanel(@"Failed to make directory", @"/%@\n%@\n%@", 
                        @"Ok", NULL, NULL,
                        [browser path], name,
                        ([last length]>80 ? 
                         [last substringToIndex:79] : last));
    }
    else{
        [browser reloadColumn:column];
        [browser scrollColumnToVisible:column];
    }

    return self;
}

- rmdir:(id)sender
{
    return [self deleteItemIsDirectory:YES];
}

- putTAR:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];

    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line, *diagnostics;

    int column = [browser selectedColumn], current;
    if([[browser selectedCell] isLeaf]==NO){
        column++;
    }

    [openPanel 
        setTitle:
            [NSString stringWithFormat:@"Upload TAR to /%@",
                      [browser path]]];
    [openPanel setAllowsMultipleSelection:NO];
    [openPanel setPrompt:@"TAR archive:"];
    [openPanel setCanChooseDirectories:NO];
    [openPanel setRequiredFileType:@"tar"];

    if([openPanel runModalForTypes:[NSArray arrayWithObject:@"tar"]]==
       NSOKButton){
        args = [self cmdOptionsForColumn:column];
        [args addObject:@"-Tx"];
        [args addObject:[openPanel filename]];
                      
        diagnostics = [NSString stringWithFormat:@"upload from %@\n",
                                [openPanel filename]];
        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:YES];
        while(line =[cmdOutput nextObject]){
            const char *cline = [line cString];
            if(current<2 || strncmp(cline, "restore", 7)){
                diagnostics =
                    [diagnostics 
                        stringByAppendingFormat:@"%s\n", cline];
            }
            current++;
        }

        ShowDiagnostics(@"Upload TAR archive", diagnostics);

        [browser reloadColumn:column];
        [browser scrollColumnToVisible:column];
    }

    return self;
}

- getTAR:(id)sender;
{
    NSSavePanel *savePanel = [NSSavePanel savePanel];
    NSString *path = 
        [[NSFileManager defaultManager] currentDirectoryPath];
    
    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line, *diagnostics,
        *item = [[browser selectedCell] stringValue];

    int column = [browser selectedColumn], current = 0;

    [savePanel 
        setTitle:
            [NSString stringWithFormat:@"Download as TAR archive /%@",
                      [browser path]]];
    [savePanel setPrompt:@"TAR archive:"];
    [savePanel setRequiredFileType:@"tar"];

    if([savePanel runModalForDirectory:path file:item]==NSOKButton){
        args = [self cmdOptionsForColumn:column];
        [args addObject:@"-Tc"];
        [args addObject:[savePanel filename]];
        if(column>2){
            [args addObject:item];
        }
                      
        diagnostics = [NSString stringWithFormat:@"download to %@\n",
                                [savePanel filename]];
        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:YES];
        while(line =[cmdOutput nextObject]){
            const char *cline = [line cString];
            if(current<2 ||
               (strncmp(cline, "                directory", 25) &&
                strstr(cline, "kb/s")==NULL)){
                diagnostics =
                    [diagnostics 
                        stringByAppendingFormat:@"%s\n", cline];
            }
            current++;
        }

        ShowDiagnostics(@"Download as TAR archive", diagnostics);
    }

    return self;
}

NSString *CollectDiagnostics(NSEnumerator *en, NSString *start)
{
    NSString *line, *diagnostics = start;
    int current = 0;

    while(line =[en nextObject]){
        if([line length]){
            diagnostics =
                [diagnostics stringByAppendingFormat:@"%@\n",
                             line];
            current++;
        }
    }

    return (current ? diagnostics : nil);
}

- mount:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];

    NSMutableArray *args;
    NSEnumerator *cmdOutput, *mountEnum;
    NSMutableArray *mounts = 
        [con mountedFSystemsOfType:nil share:nil];
    MInfo mi;
    NSValue *val;

    NSString *diagnostics;

    int column = [browser selectedColumn];

    [openPanel 
        setTitle:
            [NSString stringWithFormat:@"Mount /%@",
                      [browser path]]];
    [openPanel setAllowsMultipleSelection:NO];
    [openPanel setPrompt:@"Mount point:"];
    [openPanel setCanChooseDirectories:YES];
    [openPanel setCanChooseFiles:NO];

    if([openPanel runModalForTypes:nil]==NSOKButton){
        NSString *choice = [openPanel filename];

        mountEnum = [mounts objectEnumerator];
        while((val=[mountEnum nextObject])!=nil){
            [val getValue:&mi];
            if([mi.path isEqual:choice]){
                NSRunAlertPanel(@"Error", 
                                @"Mountpoint %@ already in use\n%@",
                                @"Ok", NULL, NULL, mi.path, mi.share);
                return self;
            }
        }

        args = [NSArray arrayWithObjects:[self fullShare], choice,
                        @"-o", 
                        [NSString stringWithFormat:@"username=%@",
                                  [con userAndPassword]],
                        nil];

        diagnostics = [NSString stringWithFormat:@"mount %@ on %@\n",
                                [self fullShare], choice];
        cmdOutput = 
            [con runCommand:[con smbmount] arguments:args 
                 input:nil progress:YES];

        if((diagnostics=CollectDiagnostics(cmdOutput, diagnostics))!=nil){
            ShowDiagnostics(@"Mount failed", diagnostics);
        }
        else{
            NSRunAlertPanel(@"Success",
                            @"%@ now mounted on %@",
                            @"Ok", NULL, NULL,
                            [self fullShare], choice);
        }
    }

    return self;
}

#define SHWIDTH 200
#define SHLOWER  50

- umount:(id)sender
{
    NSString *diagnostics;

    NSEnumerator *mountEnum;
    NSMutableArray *mounts = 
        [con mountedFSystemsOfType:@"smbfs" share:[self fullShare]];
    MInfo mi;
    NSValue *val;

    NSPanel *shPanel;
    NSRect shRect, matRect, dnFrame, cnFrame;
    unsigned int style = 
        NSTitledWindowMask | NSMiniaturizableWindowMask | 
        NSResizableWindowMask;
    NSView *contentView;

    NSButtonCell *proto;
    float cellHeight;
    NSMatrix *mat;
    NSButton *done, *cancel;
    int row, fc = [mounts count];

    NSEnumerator *cmdOutput;

    if(!fc){
        NSString 
            *title = 
            [NSString 
                stringWithFormat:@"Share %@ is not mounted",
                [self fullShare]];

        mounts = [con mountedFSystemsOfType:@"smbfs" share:nil];
        mountEnum = [mounts objectEnumerator];

        if([mounts count]){
            diagnostics = @"Mounted SAMBA shares:\n\n";
            while((val=[mountEnum nextObject])!=nil){
                [val getValue:&mi];
                diagnostics =
                    [diagnostics stringByAppendingFormat:@"%@ on %@\n", 
                                 mi.share, mi.path];
            }
        }
        else{
            diagnostics = @"There are no mounted SAMBA shares.\n";
        }

        ShowDiagnostics(title, diagnostics);
        return self;
    }

    proto = [[NSButtonCell alloc] initImageCell: nil];
    [proto setButtonType:NSRadioButton];
    [proto setBordered:NO];
    [proto setFont:[NSFont labelFontOfSize:
                               [[proto font] pointSize]+2]];
    
    cellHeight = [proto cellSize].height+5;

    shRect = NSMakeRect(200, 200, SHWIDTH, SHLOWER+cellHeight*fc);
    shPanel = [[NSPanel alloc] initWithContentRect:shRect
                               styleMask:style
                               backing:NSBackingStoreRetained
                               defer:NO];
    [shPanel setReleasedWhenClosed:YES];
    [shPanel setMinSize:shRect.size];
    [shPanel setTitle:@"Unmount SAMBA shares"];

    done = [NSButton new];

    [done setKeyEquivalent: @"\r"];
    [done setImagePosition:NSImageRight];
    [done setImage:[NSImage imageNamed: @"common_ret"]];
    [done setAlternateImage:[NSImage imageNamed: @"common_retH"]];

    [done setTitle:@"Unmount"];
    [done setTarget:self];
    [done setAction:@selector(umountModal:)];
    [done setTag:YES];
    [done sizeToFit];
    [done setAutoresizingMask:NSViewMinXMargin];

    dnFrame = [done frame];
    [done setFrame:NSMakeRect(SHWIDTH-dnFrame.size.width-10, 
                              10,
                              dnFrame.size.width, 
                              dnFrame.size.height)];

    cancel = [NSButton new];
    [cancel setTitle:@"Cancel"];
    [cancel setTarget:self];
    [cancel setAction:@selector(umountModal:)];
    [cancel setTag:NO];
    [cancel sizeToFit];
    [cancel setAutoresizingMask:NSViewMinXMargin];

    cnFrame = [cancel frame];
    [cancel setFrame:NSMakeRect(SHWIDTH-cnFrame.size.width-10
                                -dnFrame.size.width-10, 
                              10,
                              cnFrame.size.width, 
                              cnFrame.size.height)];
    
    matRect = NSMakeRect(5, SHLOWER, SHWIDTH-5, cellHeight*fc);
    mat = [[NSMatrix alloc] initWithFrame:matRect
                            mode:NSRadioModeMatrix
                            prototype:proto
                            numberOfRows:fc
                            numberOfColumns:1];
    [mat setAutoresizingMask:NSViewWidthSizable | NSViewMinYMargin];


    mountEnum = [mounts objectEnumerator]; row=0;

    while((val=[mountEnum nextObject])!=nil){
        [val getValue:&mi];
        [[mat cellAtRow:row column:0]
            setTitle:[NSString stringWithFormat:@"%@ on %@", 
                               mi.share, mi.path]];
        row++;
    }

    contentView = [shPanel contentView];
    [contentView addSubview:mat];
    [contentView addSubview:done];
    [contentView addSubview:cancel];

    [mat sizeToFit];
    matRect = [mat frame];
    if(matRect.size.width>SHWIDTH-5){
        shRect = [shPanel frame];
        shRect.size.width = matRect.size.width+20;

        [shPanel setFrame:shRect display:YES];
        [shPanel setMinSize:shRect.size];
    }

    [[NSApplication sharedApplication] runModalForWindow:shPanel];
    [shPanel close];

    if(umount==NO){
        return self;
    }

    row = [mat selectedRow];
    val = [mounts objectAtIndex:row];
    [val getValue:&mi];

    cmdOutput = 
        [con runCommand:[con smbumount] 
             arguments:[NSArray arrayWithObject:mi.path] 
             input:nil progress:YES];

    diagnostics = [NSString stringWithFormat:
                                @"unmount %@; mount point %@\n",
                            [self fullShare], mi.path];

    if((diagnostics=CollectDiagnostics(cmdOutput, diagnostics))!=nil){
        ShowDiagnostics(@"Unmount failed", diagnostics);
    }
    else{
        NSRunAlertPanel(@"Success",
                        @"umounted %@; path was %@",
                        @"Ok", NULL, NULL,
                        [self fullShare], mi.path);
    }

    return self;
}

int Newlines(NSString *msg)
{
    int lines = 0;
    const char *mstr = [msg cString];

    while(*mstr){
        if(*mstr++ == '\n'){
            lines++;
        }
    }

    return lines;
}

#define MAXLEN 1600

- (NSString *)checkLength:(NSString *)msg
{
    int len;

    if(msg==nil || ![msg length]){
        return @"Message must not be empty";
    }

    len = [msg length] + Newlines(msg);
    if(len>MAXLEN){
        NSString *emsg = @"message too long (DOS encoding)\n";
        emsg = [emsg stringByAppendingFormat:@"got %d characters\n", len];
        emsg = [emsg stringByAppendingFormat:@"max is %d", MAXLEN];
        return emsg;
    }
     
    return nil;
}

- winpopup:(id)sender
{
    NSString *host = [[browser selectedCell] stringValue];
    NSString *title, *message, *prefix;

    BOOL success = NO;

    NSArray *args;
    NSEnumerator *cmdOutput;

    NSString *line;

    title = [NSString stringWithFormat:@"Messaging %@, domain %@",
                      host, [self domain]];
    if((message = 
        GetText(title, self, @selector(checkLength:)))==nil){
        return self;
    }

    prefix = [NSString stringWithFormat:@"sent %d bytes",
                       [message length]+Newlines(message)];

    args = [NSArray arrayWithObjects:@"-M", host, 
                    @"-U", [con user], nil];
    cmdOutput = 
        [con runCommand:[con smbclient] arguments:args 
             input:message progress:NO];

    while(line = [cmdOutput nextObject]){
        last = line;
        if([line hasPrefix:@"Connected"]){
            if((line = [cmdOutput nextObject]) &&
               [line hasPrefix:prefix]){
                success = YES;
                break;
            }
            last = line;
        }
    }

    if(success==YES){
        NSRunAlertPanel(@"Success", @"%@\nmessage delivered; %@", 
                        @"Ok", NULL, NULL,
                        title, prefix);
    }
    else{
        NSRunAlertPanel(@"Error", @"%@\ngot: %@\nexpected: %@", 
                        @"Ok", NULL, NULL,
                        title, last, prefix);
    }


    return self;
}


- sprint:(id)sender
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];

    NSMutableArray *args;
    NSEnumerator *cmdOutput;

    NSString *line;

    [openPanel 
        setTitle:
            [NSString stringWithFormat:@"Print to /%@",
                      [browser pathToColumn:3]]];
    [openPanel setAllowsMultipleSelection:NO];
    [openPanel setPrompt:@"File:"];
    [openPanel setCanChooseDirectories:NO];

    if([openPanel runModalForTypes:nil]==NSOKButton){
        args = [self cmdOptionsForColumn:3];
        [args addObject:@"-c"];
        [args addObject:
                  [NSString stringWithFormat:@"print \"%@\"",
                            [openPanel filename]]];

        cmdOutput = 
            [con runCommand:[con smbclient] arguments:args 
                 input:nil progress:YES];

        while(line =[cmdOutput nextObject]){
            last = line;
        }

        if([last hasPrefix:@"putting file"]==NO){
            NSRunAlertPanel(@"Print failed", @"/%@\n%@", 
                            @"Ok", NULL, NULL,
                            [browser pathToColumn:2], last);
        }
    }

    return self;
}

- (BOOL)validateMenuItem:(id <NSMenuItem>)menuItem
{
    NSString *title = [menuItem title];
    int column = [browser selectedColumn];
    BOOL leaf = [[browser selectedCell] isLeaf];

    if(column==-1){
        return NO;
    }

    if([title isEqualToString:@"Print"]==YES && 
       (column!=2 || leaf==NO)){
        return NO;
    }
    else if([title isEqualToString:@"Upload"]==YES &&
            (column<2 || (column==2 && leaf==YES))){
        return NO;
    }
    else if([title isEqualToString:@"Download"]==YES && 
            (column<3 || leaf==NO)){
        return NO;
    }
    else if([title isEqualToString:@"Delete"]==YES && 
            (column<3 || leaf==NO)){
        return NO;
    }
    else if([title isEqualToString:@"Make directory"]==YES && 
            (column<2 || (column==2 && leaf==YES))){
        return NO;
    }
    else if([title isEqualToString:@"Delete directory"]==YES && 
            (column<3 || leaf==YES)){
        return NO;
    }
    else if([title isEqualToString:@"Upload (TAR)"]==YES &&
            (column<2 || (column==2 && leaf==YES))){
        return NO;
    }
    else if([title isEqualToString:@"Download (TAR)"]==YES && 
            (column<2 || leaf==YES)){
        return NO;
    }
    else if([title isEqualToString:@"Mount"]==YES && 
            (column!=2 || leaf==YES)){
        return NO;
    }
    else if([title isEqualToString:@"Unmount"]==YES && 
            (column!=2 || leaf==YES)){
        return NO;
    }
    else if([title isEqualToString:@"Winpopup"]==YES && 
            column!=1){
        return NO;
    }
    
    return YES;
}

- cleanup
{
    NSNotificationCenter *ncenter = 
        [NSNotificationCenter defaultCenter];

    NSLog(@"closing %@", self);
    [ncenter removeObserver:self];
    
    return self;
}

@end


