
#include "DiagnosticsPanel.h"

NSString *RunPrompt(NSString *title, NSString *pref)
{
    NSPanel *prompt;
    NSTextField *tfield;
    NSString *result = nil;
    
    NSView *sview;
    NSEnumerator *en;
    NSRect frame;

    prompt = NSGetAlertPanel(title, pref, @"OK", @"Cancel", nil);

    en = [[[prompt contentView] subviews] objectEnumerator];
    while((sview = [en nextObject])!=nil){
        if([sview isKindOfClass:[NSTextField class]] &&
           [[(NSTextField *)sview stringValue] isEqual:pref]){
            break;
        }
    }
    
    frame = [sview frame];
    tfield = [[NSTextField alloc] 
                 initWithFrame:
                     NSMakeRect(frame.origin.x+frame.size.width+20,
                                frame.origin.y,
                                240,
                                frame.size.height+5)];

    [[prompt contentView] addSubview:tfield];
    [prompt makeFirstResponder:tfield];
    [prompt makeKeyAndOrderFront:nil];

    [tfield performClick:nil];

    if([(id)prompt runModal]==NSAlertDefaultReturn){
        result = [[tfield stringValue] stringByTrimmingSpaces];
        if(result!=nil && [result isEqual:@""]==YES){
            result = nil;
        }
    }

    [tfield removeFromSuperview];
    NSReleaseAlertPanel(prompt);

    return result;
}

void ShowDiagnostics(NSString *title, NSString *diagnostics)
{
    DiagnosticsPanel *dpanel = 
        [[DiagnosticsPanel alloc]
            initWithTitle:title andContent:diagnostics];
    [[NSApplication sharedApplication]
        runModalForWindow:dpanel];
    [dpanel close];
}

@implementation DiagnosticsPanel

#define DGWIDTH 500

- initWithTitle:(NSString *)title andContent:(NSString *)diagnostics
{
    int style =  NSTitledWindowMask | NSClosableWindowMask;

    NSView *contentView;
    NSScrollView *scrollView;
    NSTextView *textView;
    NSTextField *textField;
    NSButton *button;
    NSRect frame;

    NSFont *font = [NSFont userFixedPitchFontOfSize:12];
    NSSize fdims;

    NSEnumerator *en =  
        [[diagnostics componentsSeparatedByString:@"\n"]
            objectEnumerator];
    NSString *line;
    int maxlen = -1, len, count = 0;

    while(line = [en nextObject]){
        count++;
        if((len = [line length])>maxlen){
            maxlen = len;
        }
    }
    
    textField = [[NSTextField alloc]
                    initWithFrame:
                        NSMakeRect(10, 250, DGWIDTH, 30)];
    [textField setFont:[NSFont labelFontOfSize:
                                   [[textField font] pointSize]+8]];
    [textField setStringValue:title];
    [textField setEditable:NO];
    [textField setBezeled: NO];
    [textField setDrawsBackground: NO];

    scrollView = [[NSScrollView alloc] 
                     initWithFrame:
                         NSMakeRect(10, 40, DGWIDTH, 200)];
    [scrollView setHasHorizontalScroller:YES];
    [scrollView setHasVerticalScroller:YES]; 

    fdims = NSMakeSize([font boundingRectForFont].size.width,
                       [font defaultLineHeightForFont]);
    textView = 
        [[NSTextView alloc] 
            initWithFrame:NSMakeRect(0, 0,
                                     maxlen*fdims.width, 
                                     count*fdims.height)];
    [textView setBackgroundColor:[NSColor whiteColor]];
    
    [textView setString:diagnostics];
    [textView setFont:font];
    [textView setEditable:NO];

    [scrollView setDocumentView:textView];

    button = [NSButton new];
    [button setTitle:@"Ok"];
    [button setKeyEquivalent: @"\r"];
    [button setImagePosition:NSImageRight];
    [button setImage:[NSImage imageNamed: @"common_ret"]];
    [button setAlternateImage:[NSImage imageNamed: @"common_retH"]];
    [button sizeToFit];

 
    frame = [button frame];
    [button setFrame:NSMakeRect(DGWIDTH+10-frame.size.width, 
                                10,
                                frame.size.width, 
                                frame.size.height)];

    [super
        initWithContentRect:
            NSMakeRect(0, 0, DGWIDTH+20, 300)
        styleMask:style
        backing:NSBackingStoreBuffered
        defer:NO];
    [super setReleasedWhenClosed:YES];
    [super setTitle:title];

    [button setTarget:self];
    [button setAction:@selector(done:)];
    
    contentView = [self contentView];
    [contentView addSubview:textField];
    [contentView addSubview:scrollView];
    [contentView addSubview:button];

    return self;
}

- done:(id)sender
{
    [[NSApplication sharedApplication]
        stopModal];
    return self;
}

@end
