
#include "GetText.h"


NSString *GetText(NSString *title, id vid, SEL vsel)
{
    GetTextPanel *gpanel = 
        [[GetTextPanel alloc]
            initWithTitle:title vid:vid vSEL:vsel];
    NSString *result;

    [[NSApplication sharedApplication]
        runModalForWindow:gpanel];
    result = [gpanel string];
    [gpanel close];

    return result;
}

@implementation GetTextPanel

#define GTWIDTH 500

#define INITIALW 200
#define INITIALH 200

- initWithTitle:(NSString *)title vid:(id)vid vSEL:(SEL)vsel
{
    int style =  NSTitledWindowMask | NSClosableWindowMask;

    NSView *contentView;
    NSScrollView *scrollView;
    NSTextField *textField;
    NSButton *done, *cancel;
    NSRect textRect, dframe, cframe;

    NSFont *font = [NSFont userFixedPitchFontOfSize:12];
    
    verifier = vid;
    verify = vsel;

    textField = [[NSTextField alloc]
                    initWithFrame:
                        NSMakeRect(10, 250, GTWIDTH, 30)];
    [textField setFont:[NSFont labelFontOfSize:
                                   [[textField font] pointSize]+8]];
    [textField setStringValue:title];
    [textField setEditable:NO];
    [textField setBezeled: NO];
    [textField setDrawsBackground: NO];

    scrollView = [[NSScrollView alloc] 
                     initWithFrame:
                         NSMakeRect(10, 40, GTWIDTH, 200)];
    [scrollView setHasHorizontalScroller:NO];
    [scrollView setHasVerticalScroller:YES]; 

    textRect = [[scrollView contentView] frame];
    textView = 
        [[NSTextView alloc] 
            initWithFrame:textRect];
    [textView setBackgroundColor:[NSColor whiteColor]];
    
    [textView setString:@"--------------------"]; // pick up font
    [textView setFont:font];
    [textView setEditable:YES];
    [textView setString:@""];

    [textView setVerticallyResizable:YES];
    [textView setHorizontallyResizable:YES];

    [[textView textContainer] 
        setContainerSize:NSMakeSize(textRect.size.width, 1e7)];


    [scrollView setDocumentView:textView];

    done = [NSButton new];
    [done setTitle:@"Ok"];
    [done setKeyEquivalent: @"\r"];
    [done setImagePosition:NSImageRight];
    [done setImage:[NSImage imageNamed: @"common_ret"]];
    [done setAlternateImage:[NSImage imageNamed: @"common_retH"]];
    [done sizeToFit];

    dframe = [done frame];
    [done setFrame:NSMakeRect(GTWIDTH+10-dframe.size.width, 
                              10,
                              dframe.size.width, 
                              dframe.size.height)];

    [done setTarget:self];
    [done setAction:@selector(click:)];
    [done setTag:NSAlertDefaultReturn];
    
    cancel = [NSButton new];
    [cancel setTitle:@"Cancel"];
    [cancel sizeToFit];

    cframe = [cancel frame];
    [cancel setFrame:NSMakeRect(GTWIDTH+
                                10-cframe.size.width-5-dframe.size.width,
                                10,
                                cframe.size.width, 
                                cframe.size.height)];

    [cancel setTarget:self];
    [cancel setAction:@selector(click:)];
    [cancel setTag:NSAlertAlternateReturn];
    
    [super
        initWithContentRect:
            NSMakeRect(0, 0, GTWIDTH+20, 300)
        styleMask:style
        backing:NSBackingStoreBuffered
        defer:NO];
    [super setReleasedWhenClosed:YES];
    [super setTitle:title];

    contentView = [self contentView];
    [contentView addSubview:textField];
    [contentView addSubview:scrollView];
    [contentView addSubview:done];
    [contentView addSubview:cancel];

    return self;
}

- (NSString *)string
{
    return (result==NSAlertDefaultReturn ? 
            [textView string] : nil);
}

- click:(id)sender
{
    NSString *msg = [textView string], *err;

    if((result=[sender tag])==NSAlertDefaultReturn){
        if(verifier!=nil){
            err = [verifier performSelector:verify withObject:msg];
            if(err != nil){
                NSRunAlertPanel(@"Error", err, @"Ok", NULL, NULL);
                return self;
            }
        }
    }

    [[NSApplication sharedApplication]
        stopModal];
    return self;
}

@end
